package gssim.db.mysql;

import gssim.db.dao.data.Task;
import gssim.db.find.Finder;
import gssim.db.utils.PreparedStatementAdapter;
import gssim.db.utils.ResultSetAdapter;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.logging.LogFactory;

public class TaskDAO extends AbstractDAO implements gssim.db.dao.TaskDAO {

	public enum Type {
		JOB, TASK, PROCESS_SET
	}

	private PreparedStatementAdapter findJobs;
	private PreparedStatementAdapter findTasks;
	private PreparedStatementAdapter findProcessSets;

	protected TaskDAO(Connection conn) throws SQLException {
		super(conn);
		this.findJobs = createFindJobsStmt();
		this.findTasks = createFindTasksStmt();
		this.findProcessSets = createFindProcessSetsStmt();
		log = LogFactory.getLog(TaskDAO.class);
		setRelationName("task");
	}

	@Override
	protected PreparedStatementAdapter createDeleteStmt() throws SQLException {
		String sql = "DELETE FROM task WHERE PK = ?";
		PreparedStatementAdapter stmt = new PreparedStatementAdapter(this
				.getConnection().prepareStatement(sql));
		return stmt;
	}

	@Override
	protected PreparedStatementAdapter createFindStmt() throws SQLException {

		String sql = "SELECT * FROM task WHERE (exp_id = ? OR ISNULL(?)) "
				+ "AND (job_id = ? OR ISNULL(?)) AND (task_id = ? OR "
				+ "ISNULL(?)) AND (proc_set_id = ? OR ISNULL(?)) "
				+ "AND (exec_start_date between ? AND ?) AND "
				+ "(exec_finish_date between ? AND ?)";

		PreparedStatementAdapter stmt = new PreparedStatementAdapter(this
				.getConnection().prepareStatement(sql));
		return stmt;
	}

	protected PreparedStatementAdapter createFindJobsStmt() throws SQLException {

		String sql = "SELECT * FROM task WHERE (task_id IS NULL) AND "
				+ "(proc_set_id IS NULL) AND (exp_id = ? OR ISNULL(?)) AND "
				+ "(job_id = ? OR ISNULL(?)) AND (task_id = ? OR "
				+ "ISNULL(?)) AND (proc_set_id = ? OR ISNULL(?)) "
				+ "AND (exec_start_date between ? AND ?) AND "
				+ "(exec_finish_date between ? AND ?)";

		PreparedStatementAdapter stmt = new PreparedStatementAdapter(this
				.getConnection().prepareStatement(sql));
		return stmt;
	}

	protected PreparedStatementAdapter createFindTasksStmt()
			throws SQLException {

		String sql = "SELECT * FROM task WHERE (task_id IS NOT NULL) AND "
				+ "(proc_set_id IS NULL) AND (exp_id = ? OR ISNULL(?)) AND "
				+ "(job_id = ? OR ISNULL(?)) AND (task_id = ? OR "
				+ "ISNULL(?)) AND (proc_set_id = ? OR ISNULL(?)) "
				+ "AND (exec_start_date between ? AND ?) AND "
				+ "(exec_finish_date between ? AND ?)";

		PreparedStatementAdapter stmt = new PreparedStatementAdapter(this
				.getConnection().prepareStatement(sql));
		return stmt;
	}

	protected PreparedStatementAdapter createFindProcessSetsStmt()
			throws SQLException {

		String sql = "SELECT * FROM task WHERE (task_id IS NOT NULL) AND "
				+ "(proc_set_id IS NOT NULL) AND (exp_id = ? OR ISNULL(?)) AND"
				+ " (job_id = ? OR ISNULL(?)) AND (task_id = ? OR "
				+ "ISNULL(?)) AND (proc_set_id = ? OR ISNULL(?)) "
				+ "AND (exec_start_date between ? AND ?) AND "
				+ "(exec_finish_date between ? AND ?)";

		PreparedStatementAdapter stmt = new PreparedStatementAdapter(this
				.getConnection().prepareStatement(sql));
		return stmt;
	}

	@Override
	protected PreparedStatementAdapter createInsertStmt() throws SQLException {
		String sql = "INSERT INTO task(exp_id, job_id, task_id, "
				+ "proc_set_id, parent_id, exec_start_date, exec_finish_date, "
				+ "exec_end_date, gb_submit_date, lb_submit_date, "
				+ "completion_time, exec_start_time ,execution_time, "
				+ "start_time, ready_time, flow_time, waiting_time, "
				+ "gq_waiting_time, lateness, tardiness, makespan) "
				+ "VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, "
				+ "?, ?, ?, ?, ?)";
		PreparedStatementAdapter stmt = new PreparedStatementAdapter(this
				.getConnection().prepareStatement(sql,
						Statement.RETURN_GENERATED_KEYS));

		return stmt;
	}

	@Override
	protected PreparedStatementAdapter createSelectStmt() throws SQLException {
		String sql = "SELECT * FROM task WHERE PK = ?";
		PreparedStatementAdapter stmt = new PreparedStatementAdapter(this
				.getConnection().prepareStatement(sql));
		return stmt;
	}

	@Override
	protected PreparedStatementAdapter createUpdateStmt() throws SQLException {
		String sql = "UPDATE task SET exp_id = ?, job_id = ?, task_id = ?, "
				+ "proc_set_id = ?, parent_id = ?, exec_start_date = ?, "
				+ "exec_finish_date = ?, exec_end_date = ?, gb_submit_date = ?, "
				+ "lb_submit_date = ?, completion_time = ?, exec_start_time = ? ,"
				+ "execution_time = ?, start_time = ?, ready_time = ?, "
				+ "flow_time = ?, waiting_time = ?, gq_waiting_time = ?, "
				+ "lateness = ?, tardiness = ?, makespan = ? WHERE PK = ?";
		PreparedStatementAdapter stmt = new PreparedStatementAdapter(this
				.getConnection().prepareStatement(sql));
		return stmt;
	}

	@Override
	public boolean delete(Task obj, boolean commit) throws SQLException {

		PreparedStatementAdapter stmt = getDeleteStmt();
		stmt.setInt(1, obj.getPK());

		int result = stmt.executeUpdate();

		if (result == 1) {
			if (commit)
				commit();
			logSuccess("delete");
		} else
			logError("delete", result, obj);

		return (result == 1);
	}

	@Override
	public void fillObjectFromResultSet(Task obj, ResultSet resultSet)
			throws SQLException {

		ResultSetAdapter result = (ResultSetAdapter) resultSet;

		Integer pK = result.getIntObj("PK");
		Integer expId = result.getIntObj("exp_id");
		Integer jobId = result.getIntObj("job_id");
		Integer taskId = result.getIntObj("task_id");
		Integer procSetId = result.getIntObj("proc_set_id");
		Integer parentId = result.getIntObj("parent_id");
		Timestamp execStartDate = result.getTimestamp("exec_start_date");
		Timestamp execFinishDate = result.getTimestamp("exec_finish_date");
		Timestamp execEndDate = result.getTimestamp("exec_end_date");
		Timestamp gbSubmitDate = result.getTimestamp("gb_submit_date");
		Timestamp lbSubmitDate = result.getTimestamp("lb_submit_date");
		Integer completionTime = result.getIntObj("completion_time");
		Integer execStartTime = result.getIntObj("exec_start_time");
		Integer executionTime = result.getIntObj("execution_time");
		Integer startTime = result.getIntObj("start_time");
		Integer readyTime = result.getIntObj("ready_time");
		Integer flowTime = result.getIntObj("flow_time");
		Integer waitingTime = result.getIntObj("waiting_time");
		Integer gqWaitingTime = result.getIntObj("gq_waiting_time");
		Integer lateness = result.getIntObj("lateness");
		Integer tardiness = result.getIntObj("tardiness");
		Integer makespan = result.getIntObj("makespan");

		obj.setPK(pK);
		obj.setExpId(expId);
		obj.setJobId(jobId);
		obj.setTaskId(taskId);
		obj.setProcSetId(procSetId);
		obj.setParentId(parentId);
		obj.setExecStartDate(execStartDate);
		obj.setExecFinishDate(execFinishDate);
		obj.setExecEndDate(execEndDate);
		obj.setGbSubmitDate(gbSubmitDate);
		obj.setLbSubmitDate(lbSubmitDate);
		obj.setCompletionTime(completionTime);
		obj.setExecStartTime(execStartTime);
		obj.setExecutionTime(executionTime);
		obj.setStartTime(startTime);
		obj.setReadyTime(readyTime);
		obj.setFlowTime(flowTime);
		obj.setWaitingTime(waitingTime);
		obj.setGqWaitingTime(gqWaitingTime);
		obj.setLateness(lateness);
		obj.setTardiness(tardiness);
		obj.setMakespan(makespan);

	}

	@Override
	public int fillStatementWithObject(PreparedStatement pstmt, Task obj,
			int offset) throws SQLException {

		PreparedStatementAdapter stmt = (PreparedStatementAdapter) pstmt;

		stmt.setInt(offset + 1, obj.getExpId());
		stmt.setInt(offset + 2, obj.getJobId());
		stmt.setInt(offset + 3, obj.getTaskId());
		stmt.setInt(offset + 4, obj.getProcSetId());
		stmt.setInt(offset + 5, obj.getParentId());
		stmt.setTimestamp(offset + 6, obj.getExecStartDate());
		stmt.setTimestamp(offset + 7, obj.getExecFinishDate());
		stmt.setTimestamp(offset + 8, obj.getExecEndDate());
		stmt.setTimestamp(offset + 9, obj.getGbSubmitDate());
		stmt.setTimestamp(offset + 10, obj.getLbSubmitDate());
		stmt.setInt(offset + 11, obj.getCompletionTime());
		stmt.setInt(offset + 12, obj.getExecStartTime());
		stmt.setInt(offset + 13, obj.getExecutionTime());
		stmt.setInt(offset + 14, obj.getStartTime());
		stmt.setInt(offset + 15, obj.getReadyTime());
		stmt.setInt(offset + 16, obj.getFlowTime());
		stmt.setInt(offset + 17, obj.getWaitingTime());
		stmt.setInt(offset + 18, obj.getGqWaitingTime());
		stmt.setInt(offset + 19, obj.getLateness());
		stmt.setInt(offset + 20, obj.getTardiness());
		stmt.setInt(offset + 21, obj.getMakespan());

		return offset + 21;

	}

	@Override
	public List<Task> find(Finder finder) throws SQLException {

		PreparedStatementAdapter stmt = getFindStmt();
		List<Task> list = new ArrayList<Task>();

		stmt.setInt(1, finder.getInteger("expId"));
		stmt.setInt(2, finder.getInteger("expId"));
		stmt.setInt(3, finder.getInteger("jobId"));
		stmt.setInt(4, finder.getInteger("jobId"));
		stmt.setInt(5, finder.getInteger("taskId"));
		stmt.setInt(6, finder.getInteger("taskId"));
		stmt.setInt(7, finder.getInteger("procSetId"));
		stmt.setInt(8, finder.getInteger("procSetId"));
		stmt.setTimestamp(9, finder.getMinTimestamp("minExecutionStartDate"));
		stmt.setTimestamp(10, finder.getMaxTimestamp("maxExecutionStartDate"));
		stmt.setTimestamp(11, finder.getMinTimestamp("minExecutionFinishDate"));
		stmt.setTimestamp(12, finder.getMaxTimestamp("maxExecutionFinishDate"));

		ResultSetAdapter result = new ResultSetAdapter(stmt.executeQuery());

		while (result.next()) {
			Task obj = new Task();
			fillObjectFromResultSet(obj, result);
			list.add(obj);
		}

		return list;

	}

	public List<Task> find(Finder finder, Type type) throws SQLException {

		PreparedStatementAdapter stmt = null;

		switch (type) {
		case JOB:
			stmt = getFindJobsStmt();
			break;
		case TASK:
			stmt = getFindTasksStmt();
			break;
		case PROCESS_SET:
			stmt = getFindProcessSetsStmt();
			break;
		default:
			throw new SQLException("Find statement is null");
		}

		List<Task> list = new ArrayList<Task>();

		stmt.setInt(1, finder.getInteger("expId"));
		stmt.setInt(2, finder.getInteger("expId"));
		stmt.setInt(3, finder.getInteger("jobId"));
		stmt.setInt(4, finder.getInteger("jobId"));
		stmt.setInt(5, finder.getInteger("taskId"));
		stmt.setInt(6, finder.getInteger("taskId"));
		stmt.setInt(7, finder.getInteger("procSetId"));
		stmt.setInt(8, finder.getInteger("procSetId"));
		stmt.setTimestamp(9, finder.getMinTimestamp("minExecutionStartDate"));
		stmt.setTimestamp(10, finder.getMaxTimestamp("maxExecutionStartDate"));
		stmt.setTimestamp(11, finder.getMinTimestamp("minExecutionFinishDate"));
		stmt.setTimestamp(12, finder.getMaxTimestamp("maxExecutionFinishDate"));

		ResultSetAdapter result = new ResultSetAdapter(stmt.executeQuery());

		while (result.next()) {
			Task obj = new Task();
			fillObjectFromResultSet(obj, result);
			list.add(obj);
		}

		return list;

	}

	@Override
	public int insert(Task obj, boolean commit) throws SQLException {

		PreparedStatementAdapter stmt = getInsertStmt();

		fillStatementWithObject(stmt, obj, 0);

		int result = stmt.executeUpdate();

		if (result == 1) {
			ResultSet gk = stmt.getGeneratedKeys();
			Integer pK = gk.next() ? gk.getInt(1) : -1;
			obj.setPK(pK);

			if (commit)
				commit();

			logSuccess("insert");

		} else
			logError("insert", result, obj);

		return result;
	}

	public int[] insertBatch(List<Task> objList, boolean commit)
			throws SQLException {

		PreparedStatementAdapter stmt = getInsertStmt();

		for (int i = 0; i < objList.size(); i++) {
			fillStatementWithObject(stmt, objList.get(i), 0);
			stmt.addBatch();
		}

		int[] result = stmt.executeBatch();

		stmt.clearBatch();

		int j = 0;

		while (j < result.length)
			if (result[j] == 1)
				j++;

		if (j == result.length) {

			ResultSet gk = stmt.getGeneratedKeys();
			for (int i = 0; i < objList.size(); i++) {
				Integer PK = gk.next() ? gk.getInt(1) : -1;
				objList.get(i).setPK(PK);
			}

			logSuccess("batch insert", j);

			if (commit)
				commit();

		} else
			logError("batch insert", j);

		return result;
	}

	@Override
	public Task select(Task obj) throws SQLException {

		PreparedStatementAdapter stmt = getSelectStmt();
		stmt.setInt(1, obj.getPK());
		ResultSetAdapter result = new ResultSetAdapter(stmt.executeQuery());

		result.next();

		boolean first = result.isFirst();
		boolean last = result.isLast();

		if (first && last) {

			fillObjectFromResultSet(obj, result);
			return obj;
		}

		// no user found with specified id
		if (!first && !last) {
			return null;
		}

		// no idea how this may happen (only if id is not primary key)
		throw new SQLException("Select operation in relation "
				+ getRelationName() + " returned more than one record");

	}

	@Override
	public boolean update(Task obj, boolean commit) throws SQLException {

		PreparedStatementAdapter stmt = getUpdateStmt();

		int offset = fillStatementWithObject(stmt, obj, 0);

		stmt.setInt(offset + 1, obj.getPK());

		int result = stmt.executeUpdate();

		if (result == 1) {
			if (commit)
				commit();
			logSuccess("update");
		} else
			logError("update", result, obj);

		return (result == 1);
	}

	protected PreparedStatementAdapter getFindJobsStmt() throws SQLException {
		if (this.findJobs == null)
			throw new SQLException("Find jobs statement is null.");
		return this.findJobs;
	}

	protected PreparedStatementAdapter getFindTasksStmt() throws SQLException {
		if (this.findTasks == null)
			throw new SQLException("Find tasks statement is null.");
		return this.findTasks;
	}

	protected PreparedStatementAdapter getFindProcessSetsStmt()
			throws SQLException {
		if (this.findProcessSets == null)
			throw new SQLException("Find process sets statement is null.");
		return this.findProcessSets;
	}

}
